import numpy as np
import numba.types as nt
from numba.typed import Dict, List
from numba import njit, prange, get_num_threads, get_thread_id
from .config import conditional_njit, node_type, neighbors_type, list_node_type, edge_list_dict_type, edge_list_ty, edge_type, rect_type, list_dict_rect


@conditional_njit
def from_adj_matrix_to_adj_list(A):

    """
    Create two adjacency lists given the bi-adjacency matrix.
    adj_list2 is the true adjacency list a dictionary that maps each node to a list of its neighbors.
    While adj_list is a dictionary that maps each node to a dictionary of its neighbors and their corresponding indices in the neighbor list.

    Args:
        A (np.ndarray): The bi-adjacency matrix of the bipartite graph.
    Returns:
        adj_list (Dict[int, Dict[int, int]]): A dictionary mapping each node to a dictionary of its neighbors and their corresponding indices in the neighbor list.
        adj_list2 (Dict[int, List[int]]): A dictionary mapping each node to a list of its neighbors.
    """

    na, nb = A.shape[0], A.shape[1]
    adj_list = Dict.empty(
        key_type=node_type,
        value_type=neighbors_type
    )
    
    adj_list2 = Dict.empty(
        key_type=node_type,
        value_type=list_node_type
    )

    for i in range(na):
        neighbors = Dict.empty(
            key_type=node_type,
            value_type=node_type
        )
        cont = 0
        vicini = List.empty_list(node_type)
        for j in range(nb):
            if A[i, j] != 0:
                neighbors[j + na] = cont
                vicini.append(j + na)
                cont += 1
        adj_list[i] = neighbors
        adj_list2[i] = vicini
    
    for j in range(nb):
        neighbors = Dict.empty(
            key_type=node_type,
            value_type=node_type
        )
        cont = 0
        vicini = List.empty_list(node_type)
        for i in range(na):
            if A[i, j] != 0:
                neighbors[i] = cont
                vicini.append(i)
                cont += 1
        adj_list[j + na] = neighbors
        adj_list2[j + na] = vicini


    return adj_list, adj_list2



@conditional_njit
def get_weight_dict(A):
    """
    Create a dictionary that maps each weight to a dictionary of edges with that weight and their corresponding indices in the edge list.
    Also create a dictionary that maps each weight to a list of edges with that weight.
    And two dictionaries that map each weight to a unique index and a list of unique weights.
        Args:
        A (np.ndarray): The bi-adjacency matrix of the bipartite graph.
    Returns:
    weight_dict (Dict[int, Dict[Tuple[int, int], int]]): The comlementary dictionary of weight_dict2.
    weight_dict2 (Dict[int, List[Tuple[int, int]]]): A dictionary mapping each weight to a list of edges with that weight.
    unique_weights (Dict[int, int]): A dictionary mapping each weight to a unique index.
    unique_weights2 (List[int]): A list of unique weights. The index of each weight in the list corresponds to its unique index in the unique
    """

    na, nb= A.shape[0], A.shape[1]
    weight_dict = Dict.empty(
        key_type=node_type,
        value_type= edge_list_dict_type
    )
    weight_dict2 = Dict.empty(
        key_type=node_type,
        value_type= edge_list_ty
    )
    unique_weights = Dict.empty(
        key_type=node_type,
        value_type=node_type
    )
    unique_weights2 = List.empty_list(node_type)
    for i in range(na):
        for j in range(nb):
            if A[i, j] != 0:
                weight = A[i, j]
                edge = (i, j + na)
                if weight not in unique_weights:
                    unique_weights2.append(weight)
                    unique_weights[weight] = len(unique_weights2) - 1
                if weight not in weight_dict:
                    weight_dict[weight] = Dict.empty(
                        key_type=edge_type,
                        value_type=node_type
                    )
                if weight not in weight_dict2:
                    weight_dict2[weight] = List.empty_list(edge_type)
                weight_dict2[weight].append(edge)

                weight_dict[weight][edge] = len(weight_dict2[weight]) - 1

    return weight_dict, weight_dict2, unique_weights, unique_weights2



@conditional_njit
def update_weight_dict(weight_dict, weight_dict2, unique_weights, unique_weights2, old_weight, new_weight, old_links, new_links):
    #remove old weights
    #cicle over old weights to remove them from the weight dictionaries and unique weights if necessary
    for i in range(len(old_weight)):
        weight = old_weight[i]
        edge = old_links[i]
        if weight in weight_dict:
            index = weight_dict[weight][edge]
            del weight_dict[weight][edge]
            last_edge = weight_dict2[weight][-1]
            if edge != last_edge:
                weight_dict2[weight][index] = last_edge
                weight_dict[weight][last_edge] = index
            weight_dict2[weight].pop()
            if len(weight_dict[weight]) == 0:
                del weight_dict[weight]
                del weight_dict2[weight]
                index = unique_weights[weight]
                last_weight = unique_weights2[-1]
                if weight != last_weight:
                    unique_weights2[index] = last_weight
                    unique_weights[last_weight] = index
                unique_weights2.pop()
                del unique_weights[weight]

    #add new weight and links to the weight dictionaries and unique weights if necessary
    for i in range(len(new_weight)):
        weight = new_weight[i]
        edge = new_links[i]
        if weight not in weight_dict:
            weight_dict[weight] = Dict.empty(
                key_type=edge_type,
                value_type=node_type
            )
        if weight not in weight_dict2:
            weight_dict2[weight] = List.empty_list(edge_type)
        weight_dict2[weight].append(edge)
        
        weight_dict[weight][edge] = len(weight_dict2[weight]) - 1
        if weight not in unique_weights:
            unique_weights2.append(weight)
            unique_weights[weight] = len(unique_weights2) - 1
    return weight_dict, weight_dict2, unique_weights, unique_weights2


@conditional_njit
def update_adj_list(adj_list, adj_list2, old_links, new_links):

    """
    Update the adjacency lists when an edge is added or removed.
     1. Remove old links from the adjacency lists.
     2. Add new links to the adjacency lists.
    """
    #remove the old links from the adjacency lists using swap and pop technique.
    for i in range(len(old_links)):
        edge = old_links[i]
        u = edge[0]
        v = edge[1]
        

        index_v = adj_list[u][v]  
        index_u = adj_list[v][u] 
        
        vicini_u = adj_list2[u]
        last_neighbor_u = vicini_u[-1] 
        if index_v < len(vicini_u) - 1: 
            vicini_u[index_v] = last_neighbor_u 
            adj_list[u][last_neighbor_u] = index_v
        vicini_u.pop()  
        adj_list2[u] = vicini_u
        

        vicini_v = adj_list2[v]
        last_neighbor_v = vicini_v[-1]  
        if index_u < len(vicini_v) - 1:  
            vicini_v[index_u] = last_neighbor_v  
            adj_list[v][last_neighbor_v] = index_u
        vicini_v.pop()  
        adj_list2[v] = vicini_v
        
        
        del adj_list[u][v]
        del adj_list[v][u]

    #add new links to the adjacency lists
    for i in range(len(new_links)):
        edge = new_links[i]
        u = edge[0]
        v = edge[1]
        vicini_u = adj_list2[u]
        vicini_v = adj_list2[v]
        vicini_u.append(v)
        vicini_v.append(u)
        adj_list2[u] = vicini_u
        adj_list2[v] = vicini_v
        adj_list[u][v] = len(vicini_u)-1
        adj_list[v][u] = len(vicini_v)-1

    return adj_list, adj_list2








from ortools.sat.python import cp_model
import numpy as np
class MatriceSolutionCollector(cp_model.CpSolverSolutionCallback):
    """
    Merge all the solutions found by the solver in a list of matrices.
    Each matrix is a solution that satisfies the constraints of Strength and Degree.
    """
    def __init__(self, vars_matrix, rows, cols):
        cp_model.CpSolverSolutionCallback.__init__(self)
        self.__vars_matrix = vars_matrix
        self.__rows = rows
        self.__cols = cols
        self.__solutions = []

    def on_solution_callback(self):
        current_solution = []
        for r in range(self.__rows):
            row_vals = []
            for c in range(self.__cols):
                val = self.Value(self.__vars_matrix[r][c])
                row_vals.append(val)
            current_solution.append(row_vals)
        self.__solutions.append(current_solution)

    def get_solutions(self):
        return self.__solutions

def find_configurations(
    strength_row,
    strength_colunms,
    degree_row,
    degree_columns
):
    """
    Fixing the degree and the strenght of each node find all the possible bi-adjacency matrices that satisfy these constraints.
    The number of possible configuration increase very rapdly with the size of the graph. I suggest to use this function with small graphs (less than 16 nodes).

        args:
            strength_row (List[int]): A list of integers representing the strength of each row node.
            strength_colunms (List[int]): A list of integers representing the strength of each column node.
            degree_row (List[int]): A list of integers representing the degree of each row node.
            degree_columns (List[int]): A list of integers representing the degree of each column node.
        returns:
            List[List[List[int]]]: A list of bi-adjacency matrices that satisfy the given strength and degree constraints.

    """
    n = len(strength_row)
    m = len(strength_colunms)

    # --- 1. Controlli di consistenza di base ---
    if sum(strength_row) != sum(strength_colunms):
        print("Errore: La somma totale delle Strength righe != colonna.")
        return []
    if sum(degree_row) != sum(degree_columns):
        print("Errore: La somma totale dei Degree righe != colonna.")
        return []

    model = cp_model.CpModel()

    matrix_vars = []
    is_nonzero_vars = []

    #Defining for each entry.
    #The value of an entry can't be greater then the strenght of that node
    max_global_val = max(max(strength_row), max(strength_colunms))

    for r in range(n):
        row_x = []
        row_b = []
        for c in range(m):
            
            limit_val = min(strength_row[r], strength_colunms[c])
            x_val = model.NewIntVar(0, limit_val, f'x_{r}_{c}')
            
        
            b_val = model.NewBoolVar(f'b_{r}_{c}')

 
            model.Add(x_val > 0).OnlyEnforceIf(b_val)
            model.Add(x_val == 0).OnlyEnforceIf(b_val.Not())

            row_x.append(x_val)
            row_b.append(b_val)
        matrix_vars.append(row_x)
        is_nonzero_vars.append(row_b)


    for r in range(n):
   
        model.Add(sum(matrix_vars[r]) == strength_row[r])
   
        model.Add(sum(is_nonzero_vars[r]) == degree_row[r])

    # --- 5. Vincoli sulle Colonne ---
    # Trasponiamo le matrici di variabili per iterare sulle colonne
    col_vars = [list(i) for i in zip(*matrix_vars)]
    col_bools = [list(i) for i in zip(*is_nonzero_vars)]

    for c in range(m):
        # Vincolo Strength colonna
        model.Add(sum(col_vars[c]) == strength_colunms[c])
        # Vincolo Degree colonna
        model.Add(sum(col_bools[c]) == degree_columns[c])

    # --- 6. Risoluzione ---
    solver = cp_model.CpSolver()
    solver.parameters.enumerate_all_solutions = True
    
    solution_collector = MatriceSolutionCollector(matrix_vars, n, m)
    status = solver.Solve(model, solution_collector)

    if status == cp_model.OPTIMAL or status == cp_model.FEASIBLE:
        return solution_collector.get_solutions()
    else:
        return []


import numpy as np
import numba.types as nt
from numba.typed import Dict, List
from numba import njit, prange, get_num_threads, get_thread_id
from .config import conditional_njit, node_type, neighbors_type, list_node_type, edge_list_dict_type, edge_list_ty, edge_type, rect_type, list_dict_rect
from .moves import weight_shuffling, bridge_edge_swap, simple_edge_swap
from .data_structures import from_adj_matrix_to_adj_list, get_weight_dict


@conditional_njit
def pre_compute(A):
    """
    Pre-compute data structures for the BiWES algorithm.
        args:
            A (np.ndarray): The bi-adjacency matrix of the bipartite graph.
        returns:
            A2 (np.ndarray): A copy of the input bi-adjacency matrix.
            weight_dict (Dict[int, Dict[Tuple[int, int], int]]): The comlementary dictionary of weight_dict2.
            weight_dict2 (Dict[int, List[Tuple[int, int]]]): A dictionary mapping each weight to a list of edges with that weight.
            adj_list (Dict[int, Dict[int, int]]): A dictionary mapping each node to a dictionary of its neighbors and their corresponding indices in the neighbor list.
            adj_list2 (Dict[int, List[int]]): A dictionary mapping each node to a list of its neighbors.
            unique_weights (Dict[int, int]): A dictionary mapping each weight to a unique index.
            unique_weights2 (List[int]): A list of unique weights. The index of each weight in the list corresponds to its unique index in the unique
    """
    A2 = A.copy()
    adj_list, adj_list2 = from_adj_matrix_to_adj_list(A2)
    weight_dict, weight_dict2, unique_weights, unique_weights2 = get_weight_dict(A2)
    return A2, weight_dict, weight_dict2, adj_list, adj_list2, unique_weights, unique_weights2

@conditional_njit
def BiWES(A, n_iterations):
    """
    Perform the BiWES algoritihm to randomize a bipartite graph, preserving the degree sequence and the weight sequence.
    This algorithm precompute all the necessary data structures. So if you need to perform multiple randomizations on the same graph, it is more efficient to call pre_compute once and then call the moves directly.
        args:
            A (np.ndarray): The bi-adjacency matrix of the bipartite graph.
            n_iterations (int): The number of randomization moves to perform.
        returns:
            A2 (np.ndarray): The randomized bi-adjacency matrix after performing the BiWES algorithm.
    """
    #pre compute all the necessary data structures for the BiWES algorithm
    A2, weight_dict, weight_dict2, adj_list, adj_list2, unique_weights, unique_weights2 = pre_compute(A)
    p = 1/3
    prob = np.array([p, p, p])
    #each move is selected with equal probability.
    prob = np.cumsum(prob)
    for it in range(n_iterations):
        r = np.random.rand()
        if r < prob[0]:
            A2, weight_dict, weight_dict2, unique_weights, unique_weights2 = weight_shuffling(A2, weight_dict, weight_dict2, unique_weights, unique_weights2, adj_list, adj_list2)
        elif prob[0] <= r < prob[1]:
            A2, adj_list, adj_list2, weight_dict, weight_dict2, unique_weights, unique_weights2 = simple_edge_swap(A2, adj_list, adj_list2, weight_dict, weight_dict2, unique_weights, unique_weights2)
        else:
            A2, adj_list, adj_list2, weight_dict, weight_dict2, unique_weights, unique_weights2 = bridge_edge_swap(A2, adj_list, adj_list2, weight_dict, weight_dict2, unique_weights, unique_weights2)
    return A2



@conditional_njit
def BiWES2(A2, weight_dict, weight_dict2, unique_weights, unique_weights2, adj_list, adj_list2, n_iterations):
    """
    Perform the BiWES algoritihm to randomize a bipartite graph, preserving the degree sequence and the weight sequence.
    In this case the necessary data structures are pre-computed and passed as arguments. This is useful if you need to perform multiple randomizations on the same graph, as it avoids the overhead of pre-computing the data structures multiple times.
        args:
            A2 (np.ndarray): The bi-adjacency matrix of the bipartite graph
            weight_dict (Dict[int, Dict[Tuple[int, int], int]]): The comlementary dictionary of weight_dict2.
            weight_dict2 (Dict[int, List[Tuple[int, int]]]): A dictionary mapping each weight to a list of edges with that weight.
            adj_list (Dict[int, Dict[int, int]]): A dictionary mapping each node to a dictionary of its neighbors and their corresponding indices in the neighbor list.
            adj_list2 (Dict[int, List[int]]): A dictionary mapping each node to a list of its neighbors.
            unique_weights (Dict[int, int]): A dictionary mapping each weight to a unique index.
            unique_weights2 (List[int]): A list of unique weights. The index of each weight in the list corresponds to its unique index in the unique_weights dictionary.
            n_iterations (int): The number of randomization moves to perform.
        returns:
            A2 (np.ndarray): The randomized bi-adjacency matrix after performing the BiWES algorithm.
    """
    
    p = 1/3
    prob = np.array([p, p, p])
    prob = np.cumsum(prob)

    for it in range(n_iterations):
        r = np.random.rand()
        if r < prob[0]:
            A2, weight_dict, weight_dict2, unique_weights, unique_weights2 = weight_shuffling(A2, weight_dict, weight_dict2, unique_weights, unique_weights2, adj_list, adj_list2)
        elif prob[0] <= r < prob[1]:
            A2, adj_list, adj_list2, weight_dict, weight_dict2, unique_weights, unique_weights2 = simple_edge_swap(A2, adj_list, adj_list2, weight_dict, weight_dict2, unique_weights, unique_weights2)
        else:
            A2, adj_list, adj_list2, weight_dict, weight_dict2, unique_weights, unique_weights2 = bridge_edge_swap(A2, adj_list, adj_list2, weight_dict, weight_dict2, unique_weights, unique_weights2)
    return A2


@njit(parallel=True)
def get_p_value_matrix(A, n_iterations, n_run):

    """
    Perform the BiWES algorithm multiple times to compute the probability of obersing a weight greater than or equal to the one in the original matrix A for each edge.
        args:
            A (np.ndarray): The bi-adjacency matrix of the bipartite graph.
            n_iterations (int): The number of randomization moves to perform in each simulation.
            n_run (int): The number of simulations to run.
            number_of_threads (int, optional): The number of threads to use for parallel execution. If None, it will use all available threads.
        returns:
            p_values (np.ndarray): A matrix of the same shape as A, where each entry (i, j) contains the probability of observing a weight greater than or equal to A[i, j] in the randomized graphs generated by the BiWES algorithm.

    """
    
    na, nb = A.shape
    A_obs = A.copy()
    n_threads = get_num_threads()
    
    # Divide the total number of simulations among the available threads, ensuring that any remainder is distributed among the first few threads.
    sims_per_thread = n_run // n_threads
    remainder = n_run % n_threads
    
    thread_local_counts = np.zeros((n_threads, na, nb), dtype=np.int32)

    for tid in prange(n_threads):
        n_sims = sims_per_thread + (1 if tid < remainder else 0)
        
        #each thread works on its own copy of the adjacency matrix and pre-computed data structures to avoid race conditions
        A_current = A.copy()
        A2, weight_dict, weight_dict2, adj_list, adj_list2, unique_weights, unique_weights2 = pre_compute(A)
        
       
        for sim in range(n_sims):
            
            A_current = BiWES2(A2, weight_dict, weight_dict2, unique_weights, unique_weights2, adj_list, adj_list2, n_iterations)
            
            #count how many times each edge has a weight greater than or equal to the one in the original matrix A
            for i in range(na):
                for j in range(nb):
                    if A_current[i, j] >= A_obs[i, j]:
                        thread_local_counts[tid, i, j] += 1
        
    #merge the counts from all threads
    total_counts = np.sum(thread_local_counts, axis=0)

    return total_counts / float(n_run)




from numba import njit
import numba.types as nt


USE_NJIT = True

#GLOBAL CONFIGURATION FOR NUMBA
def conditional_njit(func):
    if USE_NJIT:
        return njit(cache=True)(func)
    else:
        return func

# ==========================================
# DEFINING NUMBA TYPES THAT WILL BE USED THROUGHOUT THE CODE
# ==========================================
node_type = nt.int64
neighbors_type = nt.DictType(node_type, node_type)
edge_type = nt.UniTuple(node_type, 2)
edge_list_dict_type = nt.DictType(edge_type, node_type)
list_node_type = nt.ListType(node_type)
edge_list_ty = nt.ListType(edge_type)
rect_type = nt.UniTuple(node_type, 4)
list_dict_rect = nt.DictType(rect_type, node_type)


import numpy as np
import numba.types as nt
from numba.typed import Dict, List
from numba import njit, prange, get_num_threads, get_thread_id
from .config import conditional_njit, node_type, neighbors_type, list_node_type, edge_list_dict_type, edge_list_ty, edge_type, rect_type, list_dict_rect
from .data_structures import update_adj_list, update_weight_dict



@conditional_njit
def rectangle_random_search(A2, adj_list, adj_list2):

    """
    Randomly search for a rectangle in the bipartite graph, using a path sampling algorithm.
        args:
            A2 (np.ndarray): the bi-adjacency matrix of the bipartite graph.
            adj_list (Dict[int, Dict[int, int]]): A dictionary mapping each node to a dictionary of its neighbors and their corresponding indices in the neighbor list.
            adj_list2 (Dict[int, List[int]]): A dictionary mapping each node to a list of its neighbors.
        returns:
            rect (tuple): a tuple of the form (u1, u2, v1, v2) representing the nodes of the rectangle found, where u1 and u2 are the
    """
    
    na = A2.shape[0]
    nb = A2.shape[1]
    #random select a node finding its neighbors
    nodo = np.random.randint(0, na+nb)
    vicini = adj_list2[nodo]
    #if the node has less than 2 neighbors it can't be part of a rectangle so the searching algorithm ends.
    if len(vicini) < 2:
        return (-1,-1,-1,-1)
    else:
        #random select 2 node amog the neighbors of "nodo" and take their neighbors.
        r1 = np.random.randint(0, len(vicini))
        r2 = np.random.randint(0, len(vicini))
        while r1 == r2:   
            r2 = np.random.randint(0, len(vicini))
        nodo2 = vicini[r2]
        nodo1 = vicini[r1]
        bol = np.random.rand()
        #randomly select one of the two neighbors and find its neighbors and call it "nodo3"
        if bol < 0.5:
            vicini_lista  = adj_list2[nodo1]
            vicini_comparison = adj_list[nodo2]
        else:
            vicini_lista  = adj_list2[nodo2]
            vicini_comparison = adj_list[nodo1]
        r3 = np.random.randint(0, len(vicini_lista))
        nodo3 = vicini_lista[r3]
        #if nodo3 is a neighbor of the remaining neighbor, the rectangle is found.
        if nodo3 in vicini_comparison and nodo3 != nodo:
            if nodo < na:
                u1 = nodo
                u2 = nodo3
                v1 = nodo1
                v2 = nodo2
            else:
                u1 = nodo1
                u2 = nodo2
                v1 = nodo
                v2 = nodo3
            #adjusting the order of nodes to node to avoid bugs.
            if u1 < u2 and v1 < v2:
                rect = (u1, u2, v1, v2)
            elif u1 < u2 and v1 > v2:
                rect = (u1, u2, v2, v1)
            elif u1 > u2 and v1 < v2:
                rect = (u2, u1, v1, v2)
            else:
                rect = (u2, u1, v2, v1)
            rect = (u1, u2, v1, v2)
            return rect
        else:
            return (-1,-1,-1,-1)
        


@conditional_njit
def weight_shuffling(A2, weight_dict, weight_dict2, unique_weights, unique_weights2, adj_list, adj_list2):
    """
    Changes the weights of 4 edge od a bipartite graph presreving the topology.
    If a rectagle is found and the link's weights allow the move a delta is selected and the move is always performed.
    Otherwise the move is not performed and BiWES will return the same graph.
    Args:
        A2 (np.ndarray): the bi-adjacency matrix of the bipartite graph.
        weight_dict (Dict[int, Dict[Tuple[int, int], int]]): A dictionary mapping each weight to a dictionary of edges with that weight and their corresponding indices in the edge list.
        weight_dict2 (Dict[int, List[Tuple[int, int]]]): A dictionary mapping each weight to a list of edges with that weight.
        unique_weights (Dict[int, int]): A dictionary mapping each weight to a unique index.
        unique_weights2 (List[int]): A list of unique weights. The index of each weight in the list corresponds to its unique index in the unique_weights dictionary.
        adj_list (Dict[int, Dict[int, int]]): A dictionary mapping each node to a dictionary of its neighbors and their corresponding indices in the neighbor list.
        adj_list2 (Dict[int, List[int]]): A dictionary mapping each node to a list of its neighbors.
        

    Returns:
        np.ndarray A2: The restructured adjacency matrix of the bipartite graph
        Dict[int, Dict[Tuple[int, int], int]]: The updated weight dictionary
        Dict[int, List[Tuple[int, int]]]: The updated weight dictionary 2
        Dict[int, int]: The updated unique weights dictionary
        List[int]: The updated unique weights list
    
    Note: There is no need to update the adjacency lists because the topology of the graph is not changed, only the weights of some edges are changed.
    """


    na = A2.shape[0]

    #using 
    rectangle = rectangle_random_search(A2, adj_list, adj_list2)

    if rectangle[0] != -1:
        u1 = rectangle[0]
        u2 = rectangle[1]
        v1 = rectangle[2]
        v2 = rectangle[3]
        w11 = A2[u1, v1-na]
        w12 = A2[u1, v2-na]
        w21 = A2[u2, v1-na]
        w22 = A2[u2, v2-na]
        old_weight = np.array([w11, w22, w12, w21], dtype=np.int64)
        vet_link = [(u1, v1), (u2, v2), (u1, v2), (u2, v1)]
        vet1 = np.array([w11, w22])
        vet2 = np.array([w12, w21])
    
        

        #mi trovo tutti i possibili delta e ne scelgo uno a caso. Questo equivale a pescare dal campione dei rettangoli con tutti i possibili pesi
        #print(rectangle)
        min_aa = np.min(vet1)-1
        min_ab = (-1)*(np.min(vet2)-1)
        
        delta = np.random.randint(min_ab, min_aa+1)
        
        A2[u1, v1-na] = w11 - delta
        A2[u2, v2-na] = w22 - delta
        A2[u1, v2-na] = w12 + delta
        A2[u2, v1-na] = w21 + delta

        new_weight = np.array([A2[u1, v1-na], A2[u2, v2-na], A2[u1, v2-na], A2[u2, v1-na]], dtype=np.int64)
        weight_dict, weight_dict2, unique_weights, unique_weights2 = update_weight_dict(weight_dict, weight_dict2, unique_weights, unique_weights2, old_weight, new_weight, vet_link, vet_link)
        

    return A2, weight_dict, weight_dict2, unique_weights, unique_weights2


@conditional_njit
def simple_edge_swap(A2, adj_list, adj_list2, weight_dict, weight_dict2, unique_weights, unique_weights2):

    """
    Changes the topology of the graph by swapping two edges of the same weight.
    If two edges of the same weight are found and the swap is possible, the move is performed. Otherwise the move is not performed and BiWES will return the same graph.
    Args:
        A2 (np.ndarray): the bi-adjacency matrix of the bipartite graph.
        adj_list (Dict[int, Dict[int, int]]): A dictionary mapping each node to a dictionary of its neighbors and their corresponding indices in the neighbor list.
        adj_list2 (Dict[int, List[int]]): A dictionary mapping each node to a list of its neighbors.
        weight_dict (Dict[int, Dict[Tuple[int, int], int]]): A dictionary mapping each weight to a dictionary of edges with that weight and their corresponding indices in the edge list.
        weight_dict2 (Dict[int, List[Tuple[int, int]]]): A dictionary mapping each weight to a list of edges with that weight.
        unique_weights (Dict[int, int]): A dictionary mapping each weight to a unique index.
        unique_weights2 (List[int]): A list of unique weights. The index of each weight in the list corresponds to its unique index in the unique_weights dictionary.
    Returns:
        all the data structures updated after the move. If the move is not performed, the same data structures are returned.
    """

    #random select a weight among the unique weights in current graph
    na = A2.shape[0]
    unique_weights2_len = len(unique_weights2)
    r = np.random.randint(0, unique_weights2_len)
    weight = unique_weights2[r]
    edges = weight_dict2[weight]
    k = len(edges)
    if k < 2:
        return A2, adj_list, adj_list2, weight_dict, weight_dict2, unique_weights, unique_weights2
    #randomply select two edges with the same weight
    i = np.random.randint(0, k)
    j = np.random.randint(0, k)
    if i != j:
        edge1 = edges[i]
        edge2 = edges[j]
        u1 = edge1[0]
        u2 = edge2[0]
        v1 = edge1[1]
        v2 = edge2[1]
        #check if the swap is possible avoiding to create multiple link
        if u1 != u2 and v1 != v2 and A2[u1, v2-na] == 0 and A2[u2, v1-na] == 0:
            #perform the swap and update the data structures
            A2[u1, v1-na] = 0
            A2[u2, v2-na] = 0
            A2[u1, v2-na] = weight
            A2[u2, v1-na] = weight
            old_weight = np.array([weight, weight], dtype=np.int64)
            old_links = [(u1, v1), (u2, v2)]
            new_weight = np.array([weight, weight], dtype=np.int64)
            new_links = [(u1, v2), (u2, v1)]

            adj_list, adj_list2 = update_adj_list(adj_list, adj_list2, old_links, new_links)
            weight_dict, weight_dict2, unique_weights, unique_weights2 = update_weight_dict(weight_dict, weight_dict2, unique_weights, unique_weights2, old_weight, new_weight, old_links, new_links)

    return A2, adj_list, adj_list2, weight_dict, weight_dict2, unique_weights, unique_weights2

@conditional_njit
def research_five_nodes_structure(A2, adj_list2):
    """
    Research through a path sampling algorithm a random chain of 5 node in a graph
        args:
            A2 (np.ndarray): the bi-adjacency matrix of the bipartite graph.
            adj_list2 (Dict[int, List[int]]): A dictionary mapping each node to a list of its neighbors.
        returns:
            structure (List[int]): a list of 5 nodes representing the chain found. If the chain is not found, the list will contain less than 5 nodes.
    """

    na = A2.shape[0]
    nb = A2.shape[1]
    n = na+ nb
    structure = List.empty_list(node_type)
    old_node = -1
    for i in range(5):
        if i == 0:
            node = np.random.randint(0, n)
            vicini = adj_list2[node]
        else:
            k = len(vicini)
            if k == 0:
                break
            if k < 2 and i != 1:
                break
            elif k == 2:
                if i != 1:
                    if vicini[0] in structure and vicini[1] in structure:
                        break
                    else:
                        if vicini[0] not in structure:
                            node = vicini[0]
                        else:
                            node = vicini[1]
                    if i == 1:
                        r = np.random.randint(0, 2)
                        node = vicini[r]
            else:
                while True:
                    r = np.random.randint(0, k)
                    node2 = vicini[r]
                    if node2 != old_node:
                        node = node2
                        break
                
        if node in structure:
            break
        else:
            vicini = adj_list2[node]
            structure.append(node)
            old_node = node
    #adjusting the order of nodes to node to avoid bugs.
    if len(structure) == 5:
        if structure[0] < na:
            temp = structure[1]
            structure[1] = structure[4]
            structure[4] = temp
            temp = structure[2]
            structure[2] = structure[4]
            structure[4] = temp
        else:
            temp = structure[0]
            structure[0] = structure[1]
            structure[1] = temp
            temp = structure[3]
            structure[3] = structure[4]
            structure[4] = temp
            temp = structure[1]
            structure[1] = structure[4]
            structure[4] = temp
            temp = structure[2]
            structure[2] = structure[4]
            structure[4] = temp

    return structure


@conditional_njit
def check_five_nodes_structure(A2, structure):
    """
    Check if the structure the five nodes structure found is suitable for bridge edge swap
            args:
                A2 (np.ndarray): the bi-adjacency matrix of the bipartite graph.
                structure (List[int]): a list of 5 nodes representing the chain found. If the chain is not found, the list will contain less than 5 nodes.
            returns:
                check (bool): True if the structure is suitable for bridge edge swap, False otherwise.
    """

    na = A2.shape[0]
    check = False
    if structure[4] < na:
        u1 = structure[0]
        u2 = structure[1]
        v1 = structure[2]
        v2 = structure[3]
        u3 = structure[4]
        if A2[u1, v2-na] == 0 and A2[u2, v1-na] == 0 and A2[u1, v1-na] != A2[u2, v2-na]:
            arr = np.array([A2[u1, v1-na], A2[u2, v2-na]])
            peso_max = np.max(arr)
            if A2[u1, v1-na] == peso_max:
                if A2[u3, v2-na] > np.abs(A2[u2, v2-na] - A2[u1, v1-na]):
                    check = True
            if A2[u2, v2-na] == peso_max:
                if A2[u3, v1-na] > np.abs(A2[u2, v2-na] - A2[u1, v1-na]):
                    check = True
    if structure[4] >= na:
        u1 = structure[0]
        u2 = structure[1]
        v1 = structure[2]
        v2 = structure[3]
        v3 = structure[4]
        if A2[u1, v2-na] == 0 and A2[u2, v1-na] == 0 and A2[u1, v1-na] != A2[u2, v2-na]:
            arr = np.array([A2[u1, v1-na], A2[u2, v2-na]])
            peso_max = np.max(arr)
            if A2[u1, v1-na] == peso_max:
                if A2[u2, v3-na] > np.abs(A2[u2, v2-na] - A2[u1, v1-na]):
                    check = True
            if A2[u2, v2-na] == peso_max:
                if A2[u1, v3-na] > np.abs(A2[u2, v2-na] - A2[u1, v1-na]):
                    check = True
    return check

@conditional_njit
def bridge_edge_swap(A2, adj_list, adj_list2, weight_dict, weight_dict2, unique_weights, unique_weights2):

    """
    Changes the topology and the edge weight distribution of the graph, swapping two edge with differet weights.
    If a suitable structure is found and the swap is possible, the move is performed. Otherwise the move is not performed and BiWES will return the same graph.
    Args:
        A2 (np.ndarray): the bi-adjacency matrix of the bipartite graph.
        adj_list (Dict[int, Dict[int, int]]): A dictionary mapping each node to a dictionary of its neighbors and their corresponding indices in the neighbor list.
        adj_list2 (Dict[int, List[int]]): A dictionary mapping each node to a list of its neighbors.
        weight_dict (Dict[int, Dict[Tuple[int, int], int]]): A dictionary mapping each weight to a dictionary of edges with that weight and their corresponding indices in the edge list.
        weight_dict2 (Dict[int, List[Tuple[int, int]]]): A dictionary mapping each weight to a list of edges with that weight.
        unique_weights (Dict[int, int]): A dictionary mapping each weight to a unique index.
        unique_weights2 (List[int]): A list of unique weights. The index of each weight in the list corresponds to its unique index in the unique_weights dictionary.
    Returns:
        all the data structures updated after the move. If the move is not performed, the original data structures are returned.
    """

    na = A2.shape[0]
    
    structure = research_five_nodes_structure(A2, adj_list2)
    
    if len(structure) == 5:

        if check_five_nodes_structure(A2, structure):
            

            na = A2.shape[0]
            u1 = structure[0]
            u2 = structure[1]
            v1 = structure[2]
            v2 = structure[3]
            w11 = A2[u1, v1-na]
            w22 = A2[u2, v2-na]
            delta = np.abs(w11 - w22)
            A2[u1, v1-na] = 0
            A2[u2, v2 - na] = 0

            if structure[4] < na:
                u3 = structure[4]
                old_weights = np.array([w11, w22, A2[u3, v1 - na], A2[u3, v2 - na]], dtype=np.int64)
                A2[u1, v2 - na] = w11
                A2[u2, v1 - na] = w22
                if w11 > w22:
                    A2[u3, v1 - na] = A2[u3, v1 - na] + delta
                    A2[u3, v2 - na] = A2[u3, v2 - na] - delta
                else:
                    A2[u3, v1 - na] = A2[u3, v1 - na] - delta
                    A2[u3, v2 - na] = A2[u3, v2 - na] + delta
                old_links = [(u1, v1), (u2, v2), (u3, v1), (u3, v2)]
                new_links = [(u1, v2), (u2, v1), (u3, v1), (u3, v2)]
                new_weights = np.array([w11, w22, A2[u3, v1 - na], A2[u3, v2 - na]], dtype=np.int64)
                
                    
            if structure[4] >= na:
                v3 = structure[4]
                old_weights = np.array([w11, w22, A2[u1, v3 - na], A2[u2, v3 - na]], dtype=np.int64)
                A2[u1, v2 - na] = w22
                A2[u2, v1 - na] = w11
                if w11 > w22:
                    A2[u1, v3 - na] = A2[u1, v3 - na] + delta
                    A2[u2, v3 - na] = A2[u2, v3 - na] - delta
                else:
                    A2[u1, v3 - na] = A2[u1, v3 - na] - delta
                    A2[u2, v3 - na] = A2[u2, v3 - na] + delta
                old_links = [(u1, v1), (u2, v2), (u1, v3), (u2, v3)]
                new_links = [(u1, v2), (u2, v1), (u1, v3), (u2, v3)]
                new_weights = np.array([w22, w11, A2[u1, v3 - na], A2[u2, v3 - na]], dtype=np.int64)
                

            #update data structures
            broken_links = [(u1, v1), (u2, v2)]
            created_links = [(u1, v2), (u2, v1)]
            adj_list, adj_list2 = update_adj_list(adj_list, adj_list2, broken_links, created_links)
            weight_dict, weight_dict2, unique_weights, unique_weights2 = update_weight_dict(weight_dict, weight_dict2, unique_weights, unique_weights2, old_weights, new_weights, old_links, new_links)
    return A2, adj_list, adj_list2, weight_dict, weight_dict2, unique_weights, unique_weights2


